# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""
Avoid breaking changes to `from shrike.pipeline.pipeline_helper import AMLPipelineHelper`
"""

from .v1 import AMLPipelineHelper
