# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

EXPERIMENTAL_WARNING_MSG = (
    "EXPERIMENTAL: shrike.distributed is an incubation contrib package"
    + "that should be considered experimental and could change at any time."
)
